import itertools
import re
from copy import deepcopy

import jsonpath_rw
import mongoengine
import pytest
import bson
from bson import ObjectId

from mongoengine_migrate.actions import AlterField
from mongoengine_migrate.exceptions import SchemaError, MigrationError, InconsistencyError
from mongoengine_migrate.graph import MigrationPolicy


class TestAlterFieldCommonDbField:
    def test_forward__for_document__should_rename_field(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()
        expect = deepcopy(dump)
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for rec in parser.find(expect):
            rec.value['new_field'] = rec.value.pop('doc1_str')

        action = AlterField('Schema1Doc1', 'doc1_str', db_field='new_field')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded_document__should_rename_field(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()
        expect = deepcopy(dump)
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for rec in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            rec.value['new_embfield'] = rec.value.pop('embdoc1_str')

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str', db_field='new_embfield')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward_backward__should_rename_field_back(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()

        action = AlterField(document_type, field_name, db_field='new_field')
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == dump


class TestAlterFieldCommonRequired:
    def test_forward__for_document_when_default_is_set__should_set_to_default_value(
            self, load_fixture, test_db, dump_db
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()
        expect = deepcopy(dump)
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for rec in parser.find(expect):
            rec.value['doc1_str_empty'] = default

        action = AlterField('Schema1Doc1', 'doc1_str_empty', required=True, default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded_document__should_make_required(
            self, load_fixture, test_db, dump_db
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()
        expect = deepcopy(dump)
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for rec in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            rec.value['embdoc1_str_empty'] = default

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str_empty', required=True, default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward_backward__for_document_when_default_is_set__should_leave_set_values(
            self, load_fixture, test_db, dump_db
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()
        expect = deepcopy(dump)
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for rec in parser.find(expect):
            rec.value['doc1_str_empty'] = default

        action = AlterField('Schema1Doc1', 'doc1_str_empty', required=True, default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_forward_backward__for_embedded_when_default_is_set__should_leave_set_values(
            self, load_fixture, test_db, dump_db
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()
        expect = deepcopy(dump)
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for rec in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            rec.value['embdoc1_str_empty'] = default

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str_empty', required=True, default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__when_default_is_not_set__should_raise_error(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField(document_type, field_name, required=True, default=None)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(SchemaError):
            action.run_forward()


class TestAlterFieldCommonDefault:
    def test_forward__for_document_when_default_is_set__should_do_nothing(
            self, load_fixture, test_db, dump_db
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()

        action = AlterField('Schema1Doc1', 'doc1_str_empty', default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == dump

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward_backward__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        default = 'test!'
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()

        action = AlterField(document_type, field_name, default=default)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == dump


class TestAlterFieldCommonUnique:
    pass  # TODO


class TestAlterFieldCommonUniqueWith:
    pass  # TODO


class TestAlterFieldCommonPrimaryKey:
    def test_forward__if_field_is_filled__do_nothing(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()

        action = AlterField('Schema1Doc1', 'doc1_str', primary_key=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == dump

    def test_forward__if_field_is_partially_filled__raise_error(self, load_fixture, test_db):
        schema = load_fixture('schema1').get_schema()
        for doc in test_db['schema1_doc1'].find(limit=2):
            test_db['schema1_doc1'].update_one({'_id': doc['_id']},
                                               {'$set': {'doc1_str_empty': 'test!'}})

        action = AlterField('Schema1Doc1', 'doc1_str_empty', primary_key=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_forward()

    def test_forward_backward__should_do_nothing(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        dump = dump_db()

        action = AlterField('Schema1Doc1', 'doc1_str', primary_key=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == dump

    def test_forward__for_embedded_document__raise_error(self, load_fixture, test_db):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str', primary_key=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(SchemaError):
            action.run_forward()

    def test_backward__for_embedded_document__raise_error(self, load_fixture, test_db):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str', primary_key=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(SchemaError):
            action.run_backward()


class TestAlterFieldCommonChoices:
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str_ten'),
        ('~Schema1EmbDoc1', 'embdoc1_str_ten'),
        ('~Schema1EmbDoc2', 'embdoc2_str_ten')
    ))
    @pytest.mark.parametrize('old_choices,new_choices', (
        (None, [str(x) for x in range(11)]),  # set up
        (['doesnt', 'match'], [str(x) for x in range(11)]),  # change
        ([], [str(x) for x in range(11)]),  # change
    ))
    def test_forward__on_setup_or_change_choices_and_if_field_values_are_in_choices__do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name, old_choices, new_choices,
    ):
        schema = load_fixture('schema1').get_schema()
        schema[document_type][field_name]['choices'] = old_choices

        dump = dump_db()

        action = AlterField(document_type, field_name, choices=new_choices)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == dump

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str_ten'),
        ('~Schema1EmbDoc1', 'embdoc1_str_ten'),
        ('~Schema1EmbDoc2', 'embdoc2_str_ten')
    ))
    @pytest.mark.parametrize('old_choices,new_choices', (
        (None, ['choices', 'which', 'doesnt', 'match']),  # set up
        ([str(x) for x in range(11)], ['choices', 'which', 'doesnt', 'match'])  # change
    ))
    def test_forward__on_setup_or_change_choices_and_if_some_values_are_not_in_choices__raise_error(
            self, load_fixture, test_db, document_type, field_name, old_choices, new_choices
    ):
        schema = load_fixture('schema1').get_schema()
        schema[document_type][field_name]['choices'] = old_choices

        action = AlterField(document_type, field_name, choices=new_choices)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_forward()

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str_ten'),
        ('~Schema1EmbDoc1', 'embdoc1_str_ten'),
        ('~Schema1EmbDoc2', 'embdoc2_str_ten')
    ))
    @pytest.mark.parametrize('old_choices,new_choices', (
        (None, [str(x) for x in range(11)]),
        ([str(x) for x in range(11)], None)
    ))
    def test_forward_backward__on_setup_or_change_choices_and_values_are_in_choices__do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name, old_choices, new_choices
    ):
        schema = load_fixture('schema1').get_schema()
        schema[document_type][field_name]['choices'] = old_choices

        dump = dump_db()

        action = AlterField(document_type, field_name, choices=new_choices)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == dump

    def test_forward_backward__for_document_if_values_became_wrong_after_forward_step__raise_error(
            self, load_fixture, test_db
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('Schema1Doc1', 'doc1_str_ten', choices=None)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        for doc in test_db['schema1_doc1'].find(limit=2):
            test_db['schema1_doc1'].update_one({'_id': doc['_id']},
                                               {'$set': {'doc1_str_ten': 'out_of_choices'}})
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_backward()

    def test_forward_backward__for_embedded_document_if_values_corrupted_after_forward_step__raise_error(
            self, load_fixture, test_db
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str_ten', choices=None)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()

        # Corrupt data in db
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_str_ten'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_str_ten'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_backward()


class TestAlterFieldCommonNull:
    def test_forward__for_document_if_field_is_unset__should_set_unset_fields_to_null(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('Schema1Doc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'doc1_str_empty' not in doc.value:
                doc.value['doc1_str_empty'] = None

        action = AlterField('Schema1Doc1', 'doc1_str_empty', null=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded_if_field_is_unset__should_set_unset_fields_to_null(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'embdoc1_str_empty' not in doc.value:
                doc.value['embdoc1_str_empty'] = None

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str_empty', null=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward_backward__for_document_if_field_is_unset__should_set_unset_fields_to_null(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('Schema1Doc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'doc1_str_empty' not in doc.value:
                doc.value['doc1_str_empty'] = None

        action = AlterField('Schema1Doc1', 'doc1_str_empty', null=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_forward_backward__for_embedded_if_field_is_unset__should_set_unset_fields_to_null(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_str_empty'] = 'test!'
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'embdoc1_str_empty' not in doc.value:
                doc.value['embdoc1_str_empty'] = None

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str_empty', null=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldCommonSparse:
    pass  # TODO


class TestAlterFieldCommonTypeKey:
    pass  # TODO


class TestAlterFieldNumberMinValue:
    @pytest.mark.parametrize('db_value,min_value,expect_value', (
        (123.45, 0, 123.45),
        (2.2, 5, 5)
    ))
    @pytest.mark.parametrize('field_name', ('doc1_float', 'doc1_int_empty', 'doc1_long'))
    def test_forward__for_document__should_set_to_min_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, min_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc[field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value[field_name] = expect_value

        action = AlterField('Schema1Doc1', field_name, min_value=min_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,min_value,expect_value', (
        (123.45, 0, 123.45),
        (2.2, 5, 5)
    ))
    @pytest.mark.parametrize('field_name', ('embdoc1_float', 'embdoc1_int_empty', 'embdoc1_long'))
    def test_forward__for_embedded_document__should_set_to_min_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, min_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1'][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1'][field_name] = expect_value
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0][field_name] = expect_value

        action = AlterField('~Schema1EmbDoc1', field_name, min_value=min_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,min_value,expect_value', (
        (123.45, 0, 123.45),
        (2.2, 5, 5)
    ))
    @pytest.mark.parametrize('field_name', ('doc1_float', 'doc1_int_empty', 'doc1_long'))
    def test_forward_backward__for_document__should_leave_min_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, min_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc[field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value[field_name] = expect_value

        action = AlterField('Schema1Doc1', field_name, min_value=min_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,min_value,expect_value', (
        (123.45, 0, 123.45),
        (2.2, 5, 5)
    ))
    @pytest.mark.parametrize('field_name', ('embdoc1_float', 'embdoc1_int_empty', 'embdoc1_long'))
    def test_forward_backward__for_embedded_document__should_set_to_min_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, min_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1'][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1'][field_name] = expect_value
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0][field_name] = expect_value

        action = AlterField('~Schema1EmbDoc1', field_name, min_value=min_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldNumberMaxValue:
    @pytest.mark.parametrize('db_value,max_value,expect_value', (
        (123.45, 200, 123.45),
        (5, 2.2, 2.2)
    ))
    @pytest.mark.parametrize('field_name', ('doc1_float', 'doc1_int_empty', 'doc1_long'))
    def test_forward__for_document__should_set_to_max_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, max_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc[field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value[field_name] = expect_value

        action = AlterField('Schema1Doc1', field_name, max_value=max_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,max_value,expect_value', (
        (123.45, 200, 123.45),
        (5, 2.2, 2.2)
    ))
    @pytest.mark.parametrize('field_name', ('embdoc1_float', 'embdoc1_int_empty', 'embdoc1_long'))
    def test_forward__for_embedded_document__should_set_to_max_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, max_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1'][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1'][field_name] = expect_value
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0][field_name] = expect_value

        action = AlterField('~Schema1EmbDoc1', field_name, max_value=max_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,max_value,expect_value', (
        (123.45, 200, 123.45),
        (5, 2.2, 2.2)
    ))
    @pytest.mark.parametrize('field_name', ('doc1_float', 'doc1_int_empty', 'doc1_long'))
    def test_forward_backward__for_document__should_leave_max_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, max_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc[field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value[field_name] = expect_value

        action = AlterField('Schema1Doc1', field_name, max_value=max_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    @pytest.mark.parametrize('db_value,max_value,expect_value', (
        (123.45, 200, 123.45),
        (5, 2.2, 2.2)
    ))
    @pytest.mark.parametrize('field_name', ('embdoc1_float', 'embdoc1_int_empty', 'embdoc1_long'))
    def test_forward_backward__for_embedded_document__should_set_to_max_value_for_non_empty_fields(
            self, load_fixture, test_db, dump_db, field_name, db_value, max_value, expect_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1'][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0][field_name] = db_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1'][field_name] = expect_value
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0][field_name] = expect_value

        action = AlterField('~Schema1EmbDoc1', field_name, max_value=max_value)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldStringMaxLength:
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__if_string_length_less_max_length__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, max_length=200)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_document_if_string_length_more_max_length__should_cut_off_string(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            doc.value['doc1_str'] = 'st'

        action = AlterField('Schema1Doc1', 'doc1_str', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded_document_if_string_length_more_max_length__should_cut_off_string(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            doc.value['embdoc1_str'] = 'st'

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward_backward__if_string_length_less_max_length__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, max_length=200)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_forward_backward__for_document_if_string_length_more_max_length__should_cut_off_string(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            doc.value['doc1_str'] = 'st'

        action = AlterField('Schema1Doc1', 'doc1_str', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_forward_backward__for_embedded_if_string_length_more_max_length__should_cut_off_string(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            doc.value['embdoc1_str'] = 'st'

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_str', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldStringMinLength:
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__if_string_length_more_min_length__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, min_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__if_string_length_less_min_length__should_raise_error(
            self, load_fixture, test_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField(document_type, field_name, min_length=200)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_forward()

    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward_backward__if_string_length_more_min_length__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, min_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldStringRegex:
    @pytest.mark.parametrize('regex', (re.compile('^str'), '^str'))
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__if_field_value_match_regex__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name, regex
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, regex=regex)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('regex', (re.compile('^unknown'), '^unknown'))
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward__if_field_value_doesnt_match_regex__should_raise_error(
            self, load_fixture, test_db, document_type, field_name, regex
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField(document_type, field_name, regex=regex)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        with pytest.raises(InconsistencyError):
            action.run_forward()

    @pytest.mark.parametrize('regex', (re.compile('^str'), '^str'))
    @pytest.mark.parametrize('document_type,field_name', (
        ('Schema1Doc1', 'doc1_str'),
        ('~Schema1EmbDoc1', 'embdoc1_str'),
        ('~Schema1EmbDoc2', 'embdoc2_str')
    ))
    def test_forward_backward__if_field_value_match_regex__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name, regex
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, regex=regex)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldUrlSchemes:
    pass  # TODO


class TestAlterFieldEmailDomainWhitelist:
    pass  # TODO


class TestAlterFieldEmailAllowUTF8User:
    pass  # TODO


class TestAlterFieldEmailAllowIPDomain:
    pass  # TODO


class TestAlterFieldDecimalForceString:
    @pytest.mark.parametrize('init_value', (3.14, '3.14'))
    def test_forward__for_document__should_cast_to_string(
            self, load_fixture, test_db, dump_db, init_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_decimal'] = init_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value['doc1_decimal'] = '3.14'

        action = AlterField('Schema1Doc1', 'doc1_decimal', force_string=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value1,init_value2', ((3.14, 2.17),  ('3.14', '2.17')))
    def test_forward__for_embedded__should_cast_to_string(
            self, load_fixture, test_db, dump_db, init_value1, init_value2
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_decimal'] = init_value1
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_decimal'] = init_value2
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1']['embdoc1_decimal'] = '3.14'
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0]['embdoc1_decimal'] = '2.17'

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_decimal', force_string=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value', (3.14, '3.14'))
    def test_forward_backward__for_document__should_cast_to_string(
            self, load_fixture, test_db, dump_db, init_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_decimal'] = init_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value['doc1_decimal'] = 3.14

        action = AlterField('Schema1Doc1', 'doc1_decimal', force_string=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value1,init_value2', ((3.14, 2.17),  ('3.14', '2.17')))
    def test_forward_backward__for_embedded__should_cast_to_string(
            self, load_fixture, test_db, dump_db, init_value1, init_value2
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_decimal'] = init_value1
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_decimal'] = init_value2
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1']['embdoc1_decimal'] = 3.14
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0]['embdoc1_decimal'] = 2.17

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_decimal', force_string=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldComplexDateTimeSeparator:
    @pytest.mark.parametrize('init_value', (
        '2020.04.03.02.01.00.000000', '2020|04|03|02|01|00|000000'
    ))
    def test_forward__for_document__should_change_separator(
            self, load_fixture, test_db, dump_db, init_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_complex_datetime'] = init_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value['doc1_complex_datetime'] = '2020|04|03|02|01|00|000000'

        action = AlterField('Schema1Doc1', 'doc1_complex_datetime', separator='|')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value1,init_value2', (
        ('2020.00.01.02.03.04.000000', '2020.04.03.02.01.00.000000'),
        ('2020|00|01|02|03|04|000000', '2020|04|03|02|01|00|000000')
    ))
    def test_forward__for_embedded__should_change_separator(
            self, load_fixture, test_db, dump_db, init_value1, init_value2
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_complex_datetime'] = init_value1
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_complex_datetime'] = init_value2
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1']['embdoc1_complex_datetime'] = \
                    '2020|00|01|02|03|04|000000'
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0]['embdoc1_complex_datetime'] = \
                    '2020|04|03|02|01|00|000000'

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_complex_datetime', separator='|')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value', (
        '2020.04.03.02.01.00.000000', '2020|04|03|02|01|00|000000'
    ))
    def test_forward_backward__for_document__should_change_separator(
            self, load_fixture, test_db, dump_db, init_value
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_complex_datetime'] = init_value
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value['doc1_complex_datetime'] = '2020.04.03.02.01.00.000000'

        action = AlterField('Schema1Doc1', 'doc1_complex_datetime', separator='|')
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    @pytest.mark.parametrize('init_value1,init_value2', (
        ('2020.00.01.02.03.04.000000', '2020.04.03.02.01.00.000000'),
        ('2020|00|01|02|03|04|000000', '2020|04|03|02|01|00|000000')
    ))
    def test_forward_backward__for_embedded__should_change_separator(
            self, load_fixture, test_db, dump_db, init_value1, init_value2
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_complex_datetime'] = init_value1
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_complex_datetime'] = init_value2
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1']['embdoc1_complex_datetime'] = \
                    '2020.00.01.02.03.04.000000'
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0]['embdoc1_complex_datetime'] = \
                    '2020.04.03.02.01.00.000000'

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_complex_datetime', separator='|')
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


@pytest.mark.skipif(mongoengine.VERSION < (0, 19, 0), reason='Mongoengine>=0.19.0 is required')
class TestAlterFieldListMaxLength:
    def test_forward__for_document__should_cut_off_a_list(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if 'doc1_list' in doc.value:
                doc.value['doc1_list'] = doc.value['doc1_list'][:2]

        action = AlterField('Schema1Doc1', 'doc1_list', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded__should_cut_off_a_list(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'embdoc1_list' in doc.value:
                doc.value['embdoc1_list'] = doc.value['embdoc1_list'][:2]

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_list', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward_backward__for_document__should_cut_off_a_list(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if 'doc1_list' in doc.value:
                doc.value['doc1_list'] = doc.value['doc1_list'][:2]

        action = AlterField('Schema1Doc1', 'doc1_list', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_forward_backward__for_embedded__should_cut_off_a_list(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for doc in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'embdoc1_list' in doc.value:
                doc.value['embdoc1_list'] = doc.value['embdoc1_list'][:2]

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_list', max_length=2)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldReferenceDbref:
    def test_forward__for_document__should_convert_to_dbref(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_ref_self'] = ObjectId('000000000000000000000002')
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if 'doc1_ref_self' in doc.value:
                doc.value['doc1_ref_self'] = bson.DBRef('schema1_doc1',
                                                        ObjectId('000000000000000000000002'))

        action = AlterField('Schema1Doc1', 'doc1_ref_self', dbref=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded__should_convert_to_dbref(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000002'))
        doc['doc1_emb_embdoc1']['embdoc1_ref_doc1'] = ObjectId('000000000000000000000002')
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000002')}, doc)
        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000003'))
        doc['doc1_emblist_embdoc1'][0]['embdoc1_ref_doc1'] = ObjectId('000000000000000000000002')
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000003')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000002'):
                doc.value['doc1_emb_embdoc1']['embdoc1_ref_doc1'] = bson.DBRef(
                    'schema1_doc1', ObjectId(f'000000000000000000000002')
                )
            if doc.value['_id'] == ObjectId(f'000000000000000000000003'):
                doc.value['doc1_emblist_embdoc1'][0]['embdoc1_ref_doc1'] = bson.DBRef(
                    'schema1_doc1', ObjectId(f'000000000000000000000002')
                )

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_ref_doc1', dbref=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    @pytest.mark.parametrize('document_type,field_name', (
            ('Schema1Doc1', 'doc1_ref_self'),
            ('~Schema1EmbDoc1', 'embdoc1_ref_doc1')
    ))
    def test_forward_backward__for_document__should_do_nothing(
            self, load_fixture, test_db, dump_db, document_type, field_name
    ):
        schema = load_fixture('schema1').get_schema()

        expect = dump_db()

        action = AlterField(document_type, field_name, dbref=True)
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect


class TestAlterFieldCachedReferenceFields:
    def test_forward__for_document_when_fields_list_become_bigger__should_do_nothing(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_cachedref_self'] = {
            '_id': ObjectId('000000000000000000000002'),
            'doc1_int': 2,
            'doc1_str': '2'
        }
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()

        action = AlterField('Schema1Doc1',
                            'doc1_cachedref_self',
                            fields=['doc1_int', 'doc1_str', 'another_field'])
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_document_when_fields_list_become_smaller__should_remove_extra_fields(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        schema['Schema1Doc1']['doc1_cachedref_self']['fields'] = ['doc1_int', 'doc1_str']

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_cachedref_self'] = \
            {'_id': ObjectId('000000000000000000000002'), 'doc1_int': 2, 'doc1_str': '2'}
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if doc.value['_id'] == ObjectId(f'000000000000000000000001'):
                doc.value['doc1_cachedref_self'] = {'_id': ObjectId('000000000000000000000002'),
                                                    'doc1_int': 2}

        action = AlterField('Schema1Doc1', 'doc1_cachedref_self', fields=['doc1_int'])
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump_db() == expect

    def test_forward__for_embedded_document__forbidden_and_should_raise_error(
            self, load_fixture, test_db
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_cachedref_self', fields=['doc1_int'])

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)

    def test_forward_backward__for_document__should_remove_extra_subfields(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        doc = test_db['schema1_doc1'].find_one(ObjectId(f'000000000000000000000001'))
        doc['doc1_cachedref_self'] = {
            '_id': ObjectId('000000000000000000000002'),
            'doc1_int': 2,
            'doc1_str': '2'
        }
        test_db['schema1_doc1'].replace_one({'_id': ObjectId(f'000000000000000000000001')}, doc)

        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for doc in parser.find(expect):
            if 'doc1_cachedref_self' in doc.value:
                doc.value['doc1_cachedref_self'] = {'_id': ObjectId('000000000000000000000002'),
                                                    'doc1_int': 2}

        action = AlterField('Schema1Doc1', 'doc1_cachedref_self', fields=['doc1_int'])
        action.prepare(test_db, schema, MigrationPolicy.strict)
        action.run_forward()
        action.cleanup()
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump_db() == expect

    def test_backward__for_embedded__forbidden_and_should_raise_error(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()

        action = AlterField('~Schema1EmbDoc1', 'embdoc1_cachedref_self', fields=['doc1_int'])

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)



__all__ = [
    'check_empty_result',
    'mongo_version'
]

import functools
import logging

from pymongo.collection import Collection

from mongoengine_migrate.exceptions import InconsistencyError
from . import flags
from mongoengine_migrate.updater import DocumentUpdater, FallbackDocumentUpdater


log = logging.getLogger('mongoengine-migrate')


def check_empty_result(collection: Collection, db_field: str, find_filter: dict) -> None:
    """
    Find records in collection satisfied to a given filter expression
    and raise error if anything found
    :param collection: pymongo collection object to find in
    :param db_field: collection field name
    :param find_filter: collection.find() method filter argument
    :raises MigrationError: if any records found
    """
    bad_records = list(collection.find(find_filter, limit=3))
    if bad_records:
        examples = (
            f'{{_id: {x.get("_id", "unknown")},...{db_field}: {x.get(db_field, "unknown")}}}'
            for x in bad_records
        )
        raise InconsistencyError(f"Field {collection.name}.{db_field} in some records "
                                 f"has wrong values. First several examples: "
                                 f"{','.join(examples)}")


def mongo_version(min_version: str = None, max_version: str = None):
    """
    Decorator restrict decorated change method execution by
    MongoDB version.

    If current db version is out of specified range then instead of
    original DocumentUpdater instance, its fallback variant will be
    passed to a method
    :param min_version: Minimum MongoDB version (including)
    :param max_version: Maximum MongoDB version (excluding)
    :return:
    """
    assert min_version or max_version

    def dec(f):
        @functools.wraps(f)
        def w(*args, **kwargs):
            invalid = min_version and flags.mongo_version < min_version \
                or max_version and flags.mongo_version >= max_version

            if invalid:
                log.debug('MongoDB version is not in range (>=%s, <%s) for method %s. '
                          'Using fallback DocumentUpdater',
                          min_version, max_version, f.__name__)
                # Inject fallback updater instead of original updater
                # on 0th place (general function) or 1st (class method)
                for ind in range(2):
                    if len(args) > ind and isinstance(args[ind], DocumentUpdater):
                        args = args[:ind] + (FallbackDocumentUpdater(args[ind]),) + args[ind + 1:]
                        break
                else:
                    raise TypeError(f"Could not find DocumentUpdater in arguments of {f.__name__}")

            return f(*args, **kwargs)

        return w
    return dec



__all__ = [
    'UNSET',
    'Diff',
    'Slotinit',
    'get_closest_parent',
    'get_document_type',
    'document_type_to_class_name'
]

import inspect
from typing import Type, Iterable, Optional, NamedTuple, Any

from mongoengine import EmbeddedDocument
from mongoengine.base import BaseDocument

from .flags import EMBEDDED_DOCUMENT_NAME_PREFIX, DOCUMENT_NAME_SEPARATOR


class _Unset:
    def __str__(self):
        return 'UNSET'

    def __repr__(self):
        return '<UNSET>'

    def __bool__(self):
        return False


#: Value indicates that such schema key is unset
UNSET = _Unset()


class Diff(NamedTuple):
    """Diff of schema key values for alter methods"""
    old: Any
    new: Any
    key: str

    def __str__(self):
        return f"Diff({self.key}: {self.old}, {self.new})"

    def __repr__(self):
        return f"<Diff({self.key}: {self.old}, {self.new})>"


class Slotinit(object):
    """
    Set class __slots__ in constructor kwargs
    If slot was omitted in kwargs then its value will be taken from
    `defaults` dict (if any). Other omitted slots will remain
    uninitialized.

    Example:
        class Car(Slotinit):
            __slots__ = ('color', 'engine_power', 'bodywork_type')
            defaults = {'color': 'black', 'bodywork_type': 'sedan'}

        # Slots will be:
        # color='blue', engine_power='100hp', bodywork_type='sedan'
        my_car = Car(color='blue', engine_power='100hp')
    """

    defaults = {}
    __slots__ = ()

    def __init__(self, **kwargs):
        for slot in self.__slots__:
            if slot in kwargs:
                setattr(self, slot, kwargs[slot])
            elif slot in self.defaults:
                setattr(self, slot, self.defaults[slot])

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Slotinit):
            return False

        try:
            return set(self.__slots__) == set(other.__slots__) \
                   and all(getattr(self, slot) == getattr(other, slot) for slot in self.__slots__)
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


def get_closest_parent(target: Type, classes: Iterable[Type]) -> Type:
    """
    Find which class in given list is the closest parent to
    a target class.
    :param target: class which we are comparing of
    :param classes:
    :return: the closest parent or None if not found
    """
    target_mro = inspect.getmro(target)
    res = None
    min_distance = float('inf')
    for parent in classes:
        found = [x for x in enumerate(target_mro) if x[1] == parent]
        if found:
            distance, klass = found[0]
            if distance == 0:  # Skip if klass is target
                continue
            if distance < min_distance:
                res = klass
                min_distance = distance

    return res


def get_document_type(document_cls: Type[BaseDocument]) -> Optional[str]:
    """
    Return document type for `document_type` parameter of Action
    :param document_cls: document class
    :return: document type or None if unable to get it (if document_cls
     is abstract)
    """
    # Class name consisted of its name and all parent names separated
    # by dots, see mongoengine.base.DocumentMetaclass
    document_type = getattr(document_cls, '_class_name')
    # Replace dots on our separator since "dictdiffer" package
    # treats dot in key as dict keys path
    document_type = document_type.replace('.', DOCUMENT_NAME_SEPARATOR)
    if issubclass(document_cls, EmbeddedDocument):
        document_type = f'{EMBEDDED_DOCUMENT_NAME_PREFIX}{document_type}'

    return document_type


def document_type_to_class_name(document_type: str) -> str:
    """
    Convert document type string to class name used by mongoengine
     (`Document._class_name` property).
     E.g. '~Doc1->Doc2' to 'Doc1.Doc2'
    :param document_type:
    :return:
    """
    cls_name = document_type.replace(DOCUMENT_NAME_SEPARATOR, '.')
    emb_prefix = EMBEDDED_DOCUMENT_NAME_PREFIX
    if cls_name.startswith(emb_prefix):
        cls_name = cls_name[len(emb_prefix):]

    return cls_name



__all__ = [
    'CreateEmbedded',
    'DropEmbedded',
    'RenameEmbedded',
    'AlterEmbedded'
]

import logging

from mongoengine_migrate.flags import EMBEDDED_DOCUMENT_NAME_PREFIX
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.utils import Diff
from .base import BaseCreateDocument, BaseDropDocument, BaseRenameDocument, BaseAlterDocument
from mongoengine_migrate.updater import DocumentUpdater, ByDocContext, ByPathContext

log = logging.getLogger('mongoengine-migrate')


class CreateEmbedded(BaseCreateDocument):
    """
    Create new embedded document
    Should have the highest priority and be at top of every migration
    since fields actions could refer to this document and its schema
    representation.
    """
    priority = 4

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if not document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is not an embedded document
            return None

        return super(CreateEmbedded, cls).build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """Embedded documents are not required to do anything"""

    def run_backward(self):
        """Embedded documents are not required to do anything"""


class DropEmbedded(BaseDropDocument):
    """
    Drop embedded document
    Should have the lowest priority and be at bottom of every migration
    since fields actions could refer to this document and its schema
    representation.
    """
    priority = 18

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if not document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is not an embedded document
            return None

        return super(DropEmbedded, cls).build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """Embedded documents are not required to do anything"""

    def run_backward(self):
        """Embedded documents are not required to do anything"""


class RenameEmbedded(BaseRenameDocument):
    """
    Rename embedded document
    Should be checked before CreateEmbedded in order to detect renaming
    """
    priority = 2

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if not document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is not an embedded document
            return None

        return super(RenameEmbedded, cls).build_object(document_type, left_schema, right_schema)

    def run_forward(self):
        """Embedded documents are not required to do anything"""

    def run_backward(self):
        """Embedded documents are not required to do anything"""


class AlterEmbedded(BaseAlterDocument):
    """Alter whole embedded document changes"""
    priority = 5

    @classmethod
    def build_object(cls, document_type: str, left_schema: Schema, right_schema: Schema):
        if not document_type.startswith(EMBEDDED_DOCUMENT_NAME_PREFIX):
            # This is not an embedded document
            return None

        return super(AlterEmbedded, cls).build_object(document_type, left_schema, right_schema)

    def change_inherit(self, updater: DocumentUpdater, diff: Diff):
        """Remove '_cls' key if EmbeddedDocument becomes non-inherit,
        otherwise do nothing
        """
        def by_path(ctx: ByPathContext):
            ctx.collection.update_many(
                {ctx.filter_dotpath + '._cls': {'$exists': True}, **ctx.extra_filter},
                {'$unset': {ctx.update_dotpath + '._cls': ''}},
                array_filters=ctx.build_array_filters()
            )

        self._check_diff(diff, False, bool)
        if diff.new:
            return

        updater.update_by_path(by_path)

    def change_dynamic(self, updater: DocumentUpdater, diff: Diff):
        """If document becomes non-dynamic then remove fields which
        are not defined in mongoengine EmbeddedDocument
        """
        def by_doc(ctx: ByDocContext):
            extra_keys = ctx.document.keys() - self_schema.keys()
            if extra_keys:
                newdoc = {k: v for k, v in ctx.document.items() if k in self_schema.keys()}
                ctx.document.clear()
                ctx.document.update(newdoc)

        self._check_diff(diff, False, bool)
        if diff.new:
            return  # Nothing to do

        # Remove fields which are not in schema
        self_schema = self._run_ctx['left_schema'][self.document_type]  # type: Schema.Document
        updater.update_by_document(by_doc)



__all__ = [
    'TypeKeyRegistryItem',
    'type_key_registry',
    'add_type_key',
    'add_field_handler',
    'CONVERTION_MATRIX'
]

import decimal
import inspect
from datetime import datetime, date
from functools import partial
from typing import Dict, Type, Optional, NamedTuple

import bson
from mongoengine import fields

from . import converters


#
# Description of data kept in MongoDB fields which operated by
# appropriate mongoengine fields
#
# A field can be null which is valid value for any type
#
# Definitions (SON==dict):
# * manual ref: dict({'_id': ObjectId(), **other_keys})
#   https://docs.mongodb.com/manual/reference/database-references/#document-references
# * dynamic ref: dict({'_cls': _class_name, '_ref': DBRef({"$ref" : collection_name, "$id" : id})})
# * dynamic document: dict({'_cls': _class_name, **other_keys})
#   used for document inheritance
# * GeoJSON: dict('type': '<type>', 'coordinates': [list, with, coordinates])
#   'type' can be: Point, MultiPoint, LineString, Polygon, MultiLineString, MultiPolygon
#    https://docs.mongodb.com/manual/reference/geojson/
#    https://docs.mongodb.com/manual/geospatial-queries/
#
# ObjectIdField:
#   ObjectId
# StringField:
#   string
# URLField(StringField):
#   string, matched to url pattern which is set by field params
# EmailField(StringField):
#   string, matched to email pattern which is set by field params
# IntField:
#   32-bit integer
#   Acceptable: 64-bit integer
#   Acceptable: double (mongo shell treats integer as double by default)
# LongField:
#   64-bit integer
#   Acceptable: 32-bit integer
#   Acceptable: double (mongo shell treats integer as double by default)
# FloatField:
#   double
#   Acceptable: 32-bit integer
#   Acceptable: 64-bit integer
# DecimalField:
#   double (if force_string=False)
#   string (string representation with dot, if force_string=True)
#   Acceptable: 32-bit integer
#   Acceptable: 64-bit integer
# BooleanField:
#   boolean
# DateTimeField:
#   date (ISODate)
#   Acceptable: string with datetime
# DateField(DateTimeField):
#   date (ISODate)
#   Acceptable: string with datetime
# ComplexDateTimeField(StringField)
#   string with datetime which is set by field params
# EmbeddedDocumentField:
#   dict (any keys without '_id')
#   dynamic document (any keys without '_id')
# GenericEmbeddedDocumentField:
#   dynamic document
# DynamicField:
#   any type (including dynamic ref, dynamic document, manual ref.
#     Embedded documents always are kept as dynamic document)
# ListField:
#   list
# EmbeddedDocumentListField(ListField)
#   list
# SortedListField(ListField)
#   list
# DictField
#   dict
# MapField(DictField):
#   dict (with restriction to dict values type)
# ReferenceField:
#   ObjectId (if dbref=False)
#   DBRef (if dbref=True)
#   DBRef with `cls` (without underscore) (if dbref=True)
# CachedReferenceField:
#   manual ref
# GenericReferenceField:
#   ObjectId
#   DBRef
#   dynamic ref
# BinaryField:
#   Binary
# FileField:
#   ObjectId (id to `files` collection entry)
# ImageField(FileField):
#   ObjectId (id to `files` collection entry)
# SequenceField (virtual field, it points to a collection with counters)
#   nothing
# UUIDField:
#   Binary (binary=True)
#   string with uuid (binary=False)
# GeoPointField:
#   2-element list with doubles or 32-bit integers
# PointField:
#   GeoJSON(Point)
# LineStringField:
#   GeoJSON(LineString)
# PolygonField:
#   GeoJSON(Polygon)
# MultiPointField:
#   GeoJSON(MultiPoint)
# MultiLineStringField:
#   GeoJSON(MultiLineString)
# MultiPolygonField:
#   GeoJSON(MultiPolygon)
# LazyReferenceField:
#   ObjectId (if dbref=False)
#   DBRef (if dbref=True)
# GenericLazyReferenceField(GenericReferenceField):
#   ObjectId
#   DBRef
#   dynamic ref


class TypeKeyRegistryItem(NamedTuple):
    """
    Information of classes for a type key
    """
    field_cls: Type[fields.BaseField]
    field_handler_cls: Optional[Type['CommonFieldHandler']]


#: Registry of available `type_key` values which can be used in schema.
#: And appropriate mongoengine field class and handler class
#:
#: This registry is filled out with all available mongoengine field
#: classes. Type key is a name of mongoengine field class. As a handler
#: there is used handler associated with this field or CommonFieldHander
type_key_registry: Dict[str, TypeKeyRegistryItem] = {}


def add_type_key(field_cls: Type[fields.BaseField]):
    """
    Add mongoengine field to type_key registry
    :param field_cls: mongoengine field class
    :return:
    """
    assert inspect.isclass(field_cls) and issubclass(field_cls, fields.BaseField), \
        f'{field_cls!r} is not a class derived from BaseField'

    type_key_registry[field_cls.__name__] = TypeKeyRegistryItem(field_cls=field_cls,
                                                                field_handler_cls=None)


def add_field_handler(field_cls: Type[fields.BaseField], handler_cls: Type['CommonFieldHandler']):
    """
    Add field handler to the type_key registry for a given field
    :param field_cls:
    :param handler_cls:
    :return:
    """
    if field_cls.__name__ not in type_key_registry:
        raise ValueError(f'Could not add handler {handler_cls!r} for unknown mongoengine field '
                         f'class {field_cls!r}')

    # Handlers can be added in any order
    # So set a handler only on those registry items where no handler
    # was set or where handler is a base class of given one
    for type_key, registry_item in type_key_registry.items():
        current_handler = registry_item.field_handler_cls
        if current_handler is None or issubclass(registry_item.field_cls, field_cls):
            type_key_registry[type_key] = TypeKeyRegistryItem(
                field_cls=registry_item.field_cls,
                field_handler_cls=handler_cls
            )


# Fill out the type key registry with all mongoengine fields
for name, member in inspect.getmembers(fields):
    if not inspect.isclass(member) or not issubclass(member, fields.BaseField):
        continue

    add_type_key(member)


DENY_BASES = {
    # Do not set converters for BooleanField, DynamicField,
    # SequenceField since they are set below
    fields.ObjectIdField: converters.deny,
    fields.StringField: converters.deny,   # + URLField, EmailField, ComplexDateTimeField
    fields.IntField: converters.deny,
    fields.LongField: converters.deny,
    fields.FloatField: converters.deny,
    fields.DecimalField: converters.deny,
    fields.DateTimeField: converters.deny,  # + DateField, ComplexDateTimeField
    fields.EmbeddedDocumentField: converters.deny,
    fields.GenericEmbeddedDocumentField: converters.deny,
    fields.ListField: converters.deny,  # + EmbeddedDocumentListField, SortedListField
    fields.DictField: converters.deny,  # + MapField
    fields.ReferenceField: converters.deny,
    fields.CachedReferenceField: converters.deny,
    fields.GenericReferenceField: converters.deny,  # + GenericLazyReferenceField
    fields.BinaryField: converters.deny,
    fields.FileField: converters.deny,  # + ImageField
    fields.UUIDField: converters.deny,
    fields.GeoPointField: converters.deny,
    fields.GeoJsonBaseField: converters.deny,  # All geo fields except GeoPointField
    fields.LazyReferenceField: converters.deny,
}


COMMON_CONVERTERS = {
    **DENY_BASES,
    fields.StringField: converters.to_string,
    fields.IntField: converters.to_int,
    fields.LongField: converters.to_long,
    fields.FloatField: converters.to_double,
    fields.DecimalField: converters.to_decimal,
    fields.DateTimeField: converters.to_date,  # + DateField
    fields.ComplexDateTimeField: converters.to_complex_datetime,
    fields.ListField: converters.item_to_list,  # + SortedListField
    fields.EmbeddedDocumentListField: converters.deny,  # Override ListField converter
}

OBJECTID_CONVERTERS = {
    **DENY_BASES,
    fields.ObjectIdField: converters.to_object_id,
    fields.StringField: converters.to_string,
    fields.URLField: converters.deny,  # Override StringField converter
    fields.EmailField: converters.deny,  # Override StringField converter
    fields.ComplexDateTimeField: converters.deny,  # Override StringField converter
    fields.ListField: converters.item_to_list,
    fields.ReferenceField: converters.to_dbref,
    fields.CachedReferenceField: converters.to_manual_ref,
    fields.FileField: converters.to_object_id,
    fields.LazyReferenceField: converters.to_dbref,
    fields.GenericReferenceField: converters.to_dynamic_ref  # + GenericLazyReferenceField
}


def get_geojson_converters(from_type):
    return {
        **DENY_BASES,
        fields.ListField: converters.item_to_list,
        fields.EmbeddedDocumentListField: converters.deny,  # Override ListField converter
        fields.DictField: converters.nothing,  # + MapField
        fields.GeoPointField: partial(converters.geojson_to_legacy_pairs, from_type=from_type),
        fields.PointField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='Point'
        ),
        fields.LineStringField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='LineString'
        ),
        fields.PolygonField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='Polygon'
        ),
        fields.MultiPointField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='MultiPoint'
        ),
        fields.MultiLineStringField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='MultiLineString'
        ),
        fields.MultiPolygonField: partial(
            converters.convert_geojson,
            from_type=from_type,
            to_type='MultiPolygon'
        ),
    }


#: Field type convertion matrix
#: Contains information which converter must be used to convert one
#: mongoengine field type to another. Used by field handlers when
#: 'type_key' schema parameter changes
#:
#: Types are searched here either as exact class equality or the nearest
#: parent. If type pair is not exists in matrix, then such convertion
#: is denied.
#:
#: Deny convertion from unknown general dicts to embedded documents,
#: manual ref, dynamic document, geojson fields because these dicts
#: have unknown structure. Also deny convertion general lists to
#: legacy geo fields, embedded document list fields for the same reason
#:
#: Format: {field_type1: {field_type2: converter_function, ...}, ...}
#:
CONVERTION_MATRIX = {
    fields.ObjectIdField: OBJECTID_CONVERTERS.copy(),
    fields.StringField: {
        **COMMON_CONVERTERS,
        fields.ObjectIdField: converters.to_object_id,
        fields.URLField: partial(converters.to_url_string, check_only=True),
        fields.EmailField: partial(converters.to_email_string, check_only=True),
        fields.ReferenceField: converters.to_dbref,
        fields.GenericReferenceField: converters.to_dbref,  # + GenericLazyReferenceField
        fields.FileField: converters.to_object_id,  # + ImageField
        fields.UUIDField: converters.to_uuid_str,
        fields.LazyReferenceField: converters.to_dbref,
    },
    fields.IntField: COMMON_CONVERTERS.copy(),
    fields.LongField: COMMON_CONVERTERS.copy(),
    fields.FloatField: COMMON_CONVERTERS.copy(),
    fields.DecimalField: COMMON_CONVERTERS.copy(),
    fields.BooleanField: {
        **COMMON_CONVERTERS,
        fields.DateTimeField: converters.deny,
        fields.ComplexDateTimeField: converters.deny
    },
    fields.DateTimeField: COMMON_CONVERTERS.copy(),

    fields.EmbeddedDocumentField: {
        # Forbid convertion to DynamicField since it requires
        # to be dynamic document (with '_cls' dict key) if it contains
        # embedded document
        **DENY_BASES,
        fields.DynamicField: converters.deny,
        fields.ListField: converters.item_to_list,
        fields.EmbeddedDocumentListField: converters.item_to_list,
        fields.DictField: converters.nothing,  # + MapField
        fields.ReferenceField: converters.deny,  # TODO: convert reference to embedded
    },
    fields.GenericEmbeddedDocumentField: {
        **DENY_BASES,
        fields.EmbeddedDocumentField: converters.remove_cls_key,
        fields.ListField: converters.item_to_list,
        fields.EmbeddedDocumentListField: partial(converters.item_to_list, remove_cls_key=True),
        fields.DictField: converters.nothing,  # + MapField
    },

    fields.DynamicField: {
        **COMMON_CONVERTERS,
        fields.ObjectIdField: converters.to_object_id,
        fields.URLField: converters.to_url_string,
        fields.EmailField: converters.to_email_string,
        fields.ComplexDateTimeField: converters.to_complex_datetime,
        fields.DictField: converters.to_object,  # + MapField
        fields.ReferenceField: converters.to_dbref,
        fields.GenericReferenceField: converters.to_dbref,  # + GenericLazyReferenceField
        fields.BinaryField: converters.to_binary,
        fields.FileField: converters.to_object_id,
        fields.UUIDField: converters.to_uuid_str,
        fields.LazyReferenceField: converters.to_dbref,
    },
    fields.ListField: {
        **DENY_BASES,
        fields.ObjectIdField: partial(converters.extract_from_list, bson.ObjectId),
        fields.StringField: partial(converters.extract_from_list, str),
        fields.IntField: partial(converters.extract_from_list, int),
        fields.LongField: partial(converters.extract_from_list, int),  # or bson.Int64
        fields.FloatField: partial(converters.extract_from_list, float),
        fields.DecimalField: partial(converters.extract_from_list, decimal.Decimal),
        fields.DateTimeField: partial(converters.extract_from_list, (datetime, date)),
        # Override ListField
        fields.EmbeddedDocumentListField: converters.deny,
        fields.DictField: partial(converters.extract_from_list, dict),  # + MapField
        fields.ReferenceField: partial(converters.extract_from_list, (bson.ObjectId, bson.DBRef)),
        fields.BinaryField: partial(converters.extract_from_list, bson.Binary),
        fields.FileField: partial(converters.extract_from_list, bson.ObjectId),
        fields.LazyReferenceField: partial(converters.extract_from_list,
                                           (bson.ObjectId, bson.DBRef)),
    },
    fields.EmbeddedDocumentListField: {
        **DENY_BASES,
        fields.EmbeddedDocumentField: partial(converters.extract_from_list, dict),
        fields.ListField: converters.nothing,
        fields.DictField: partial(converters.extract_from_list, dict),  # + MapField
    },
    # Forbid convertion for DictField almost everywhere because
    # typically a dict has unknown structure
    fields.DictField: {
        **DENY_BASES,
        fields.ListField: converters.item_to_list,
        # Override ListField
        fields.EmbeddedDocumentListField: converters.deny,
    },

    fields.ReferenceField: OBJECTID_CONVERTERS.copy(),
    fields.CachedReferenceField: {
        **DENY_BASES,
        fields.ObjectIdField: converters.to_object_id,
        fields.ListField: converters.item_to_list,
        # Override ListField
        fields.EmbeddedDocumentListField: converters.deny,
        fields.DictField: converters.nothing,  # + MapField
        fields.ReferenceField: converters.to_dbref,
        fields.GenericReferenceField: converters.to_dbref,  # + GenericLazyReferenceField
        fields.FileField: converters.to_object_id,
        fields.LazyReferenceField: converters.to_dbref,
    },
    fields.GenericReferenceField: OBJECTID_CONVERTERS.copy(),  # + GenericLazyReferenceField

    # BinaryField has unknown binary data which more likely
    # will not be decoded into one of value
    # User can write his own action using RunPython to convert it
    fields.BinaryField: {
        **DENY_BASES,
    },
    fields.FileField: OBJECTID_CONVERTERS.copy(),
    # Sequence field just points to another counter field, so do nothing
    fields.SequenceField: {
        fields.BaseField: converters.nothing
    },
    fields.UUIDField: {
        **DENY_BASES,
        fields.StringField: converters.to_string,
        fields.URLField: converters.deny,  # Override StringField converter
        fields.EmailField: converters.deny,  # Override StringField converter
        fields.ComplexDateTimeField: converters.deny,  # Override StringField converter
        fields.ListField: converters.item_to_list,
        fields.EmbeddedDocumentListField: converters.deny,   # Override ListField converter
        fields.BinaryField: converters.to_uuid_bin,
    },

    # Geo fields
    fields.GeoPointField: {
        **DENY_BASES,
        fields.ListField: converters.item_to_list,
        fields.EmbeddedDocumentListField: converters.deny,   # Override ListField converter
        fields.DictField: converters.nothing,
        fields.PointField: partial(converters.legacy_pairs_to_geojson, to_type='Point'),
        fields.LineStringField: partial(converters.legacy_pairs_to_geojson, to_type='LineString'),
        fields.PolygonField: partial(converters.legacy_pairs_to_geojson, to_type='Polygon'),
        fields.MultiPointField: partial(converters.legacy_pairs_to_geojson, to_type='MultiPoint'),
        fields.MultiLineStringField: partial(converters.legacy_pairs_to_geojson,
                                             to_type='MultiLineString'),
        fields.MultiPolygonField: partial(converters.legacy_pairs_to_geojson,
                                          to_type='MultiPolygon'),
    },
    fields.PointField: get_geojson_converters('Point'),
    fields.LineStringField: get_geojson_converters('LineString'),
    fields.PolygonField: get_geojson_converters('Polygon'),
    fields.MultiPointField: get_geojson_converters('MultiPoint'),
    fields.MultiLineStringField: get_geojson_converters('MultiLineString'),
    fields.MultiPolygonField: get_geojson_converters('MultiPolygonField'),

    fields.LazyReferenceField: OBJECTID_CONVERTERS.copy(),

    # Leave field as is if field type is unknown
    fields.BaseField: {}
}


for klass, converters_mapping in CONVERTION_MATRIX.items():
    # Add fallback converter for unknown fields
    CONVERTION_MATRIX[klass].setdefault(fields.BaseField, converters.nothing)

    # Add boolean converter for all fields
    # Any db value can get casted to boolean
    CONVERTION_MATRIX[klass].setdefault(fields.BooleanField, converters.to_bool)

    # Drop field during convertion to SequenceField
    CONVERTION_MATRIX[klass].setdefault(fields.SequenceField, converters.drop_field)

    # Add DynamicField converter for all fields since it can keep any
    # type of value
    CONVERTION_MATRIX[klass].setdefault(fields.DynamicField, converters.nothing)

    # Force set convertion between class and its parent/child class
    CONVERTION_MATRIX[klass][klass] = converters.nothing



__all__ = [
    'nothing',
    'deny',
    'drop_field',
    'item_to_list',
    'extract_from_list',
    'remove_cls_key',
    'to_manual_ref',
    'to_dbref',
    'to_dynamic_ref',
    'to_string',
    'to_int',
    'to_long',
    'to_double',
    'to_decimal',
    'to_date',
    'to_bool',
    'to_binary',
    'to_object',
    'to_object_id',
    'to_uuid_str',
    'to_uuid_bin',
    'to_url_string',
    'to_email_string',
    'to_complex_datetime',
]

import re
import uuid

import bson
from dateutil.parser import parse as dateutil_parse

from mongoengine_migrate.exceptions import MigrationError, InconsistencyError
from mongoengine_migrate.mongo import (
    check_empty_result
)
from mongoengine_migrate.updater import ByPathContext, ByDocContext, DocumentUpdater


def nothing(*args, **kwargs):
    """Converter which does nothing"""
    pass


def deny(updater: DocumentUpdater):
    """Convertion is denied"""
    raise MigrationError(f"Such type_key convertion for field "
                         f"{updater.document_type}.{updater.field_name} is forbidden")


def drop_field(updater: DocumentUpdater):
    """Drop field"""
    def by_path(ctx: ByPathContext):
        ctx.collection.update_many(
            {ctx.filter_dotpath: {'$exists': True}, **ctx.extra_filter},
            {'$unset': {ctx.update_dotpath: ''}},
            array_filters=ctx.build_array_filters()
        )

    updater.update_by_path(by_path)


def item_to_list(updater: DocumentUpdater, remove_cls_key=False):
    """Make a list with single element from every non-array value"""
    def by_doc(ctx: ByDocContext):
        if updater.field_name in ctx.document:
            f = ctx.document[updater.field_name]
            if f is not None:
                if remove_cls_key and isinstance(f, dict) and '_cls' in f:
                    del f['_cls']
                if not isinstance(f, (list, tuple)):
                    ctx.document[updater.field_name] = [f]
            else:
                ctx.document[updater.field_name] = []  # null -> []

    updater.update_by_document(by_doc)


def extract_from_list(updater: DocumentUpdater, item_type, remove_cls_key=False):
    """
    Replace every list which was met with its first element with
    checking item type. If type is other than `item_type` then
    the error will be raised
    :param updater:
    :param item_type: python type(s) to check the element
    :param remove_cls_key: if True then '_cls' keys will be removed
     from dict items if any
    :return:
    """
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        if isinstance(f, (list, tuple)):
            if f:
                f = f[0]
                if remove_cls_key and isinstance(f, dict) and '_cls' in f:
                    del f['_cls']
                if not isinstance(f, item_type) and updater.migration_policy.name == 'strict':
                    raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                             f"(should be {item_type}) in record {doc}")
            else:
                f = None
            doc[updater.field_name] = f
        elif f is not None and updater.migration_policy.name == 'strict':
            raise MigrationError(f'Could not extract item from non-list value '
                                 f'{updater.field_name}: {doc[updater.field_name]}')

    updater.update_by_document(by_doc)


def remove_cls_key(updater: DocumentUpdater):
    """Unset '_cls' key in documents if any"""
    def by_path(ctx: ByPathContext):
        ctx.collection.update_many(
            {ctx.filter_dotpath + '._cls': {'$exists': True}, **ctx.extra_filter},
            {'$unset': {ctx.update_dotpath + '._cls': ''}},
            array_filters=ctx.build_array_filters()
        )

    updater.update_by_path(by_path)


def to_object_id(updater: DocumentUpdater):
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        is_dict = isinstance(f, dict)
        if isinstance(f, str):  # ObjectId as string
            try:
                f = bson.ObjectId(f)
            except bson.errors.BSONError:
                pass

        if isinstance(f, bson.ObjectId):
            return
        elif is_dict and isinstance(f.get('_ref'), bson.DBRef):  # Already dynamic ref
            doc[updater.field_name] = f['_ref'].id
        elif isinstance(f, bson.DBRef):
            doc[updater.field_name] = f.id
        elif is_dict and isinstance(f.get('_id'), bson.ObjectId):  # manual ref
            doc[updater.field_name] = f['_id']

        elif updater.migration_policy.name == 'strict':  # Other data type
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be DBRef, ObjectId, manual ref, dynamic ref, "
                                     f"ObjectId string) in record {doc}")

    # TODO: precheck if field actually contains value other than ObjectId
    updater.update_by_document(by_doc)


def to_manual_ref(updater: DocumentUpdater):
    """Convert references (ObjectId, DBRef, dynamic ref) to manual ref
    """
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        if isinstance(f, str):  # ObjectId as string
            try:
                f = bson.ObjectId(f)
            except bson.errors.BSONError:
                pass

        is_dict = isinstance(f, dict)
        if is_dict and isinstance(f.get('_id'), bson.ObjectId):  # Already manual ref
            return
        elif is_dict and isinstance(f.get('_ref'), bson.DBRef):  # dynamic ref
            doc[updater.field_name] = {'_id': f['_ref'].id}
        elif isinstance(f, bson.DBRef):
            doc[updater.field_name] = {'_id': f.id}
        elif isinstance(f, bson.ObjectId):
            doc[updater.field_name] = {'_id': f}
        elif updater.migration_policy.name == 'strict':  # Other data type
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be DBRef, ObjectId, manual ref, dynamic ref, "
                                     f"ObjectId string) in record {doc}")

    # TODO: precheck if field actually contains value other than manual ref
    updater.update_by_document(by_doc)


def to_dbref(updater: DocumentUpdater):
    """Convert references (ObjectId, manual ref, dynamic ref) to dbref
    """
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        if isinstance(f, bson.DBRef):  # Already DBRef
            return
        elif isinstance(f, str):  # ObjectId as string
            try:
                f = bson.ObjectId(f)
            except bson.errors.BSONError:
                pass

        collection_name = ctx.collection.name if ctx.collection else None
        is_dict = isinstance(f, dict)
        if is_dict and isinstance(f.get('_id'), bson.ObjectId):  # manual ref
            doc[updater.field_name] = bson.DBRef(collection_name, f['_id'])
        elif is_dict and isinstance(f.get('_ref'), bson.DBRef):  # dynamic ref
            doc[updater.field_name] = f['_ref']
        elif isinstance(f, bson.ObjectId):
            doc[updater.field_name] = bson.DBRef(collection_name, f)
        elif updater.migration_policy.name == 'strict':  # Other data type
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be DBRef, ObjectId, manual ref, dynamic ref, "
                                     f"ObjectId string) in record {doc}")

    # TODO: precheck if field actually contains value other than DBRef
    updater.update_by_document(by_doc)


def to_dynamic_ref(updater: DocumentUpdater):
    """Convert references (ObjectId, DBRef, manual ref) to dynamic ref
    """
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        is_dict = isinstance(f, dict)
        collection_name = ctx.collection.name if ctx.collection else None

        if isinstance(f, str):  # ObjectId as string
            try:
                f = bson.ObjectId(f)
            except bson.errors.BSONError:
                pass

        # We cannot get dynamic ref from other types of refs because
        # of lack of '_cls' value. Mongoengine fields which use this
        # converter can keep DBRef. So return DBRef instead
        if is_dict and isinstance(f.get('_ref'), bson.DBRef):  # Already dynamic ref
            return
        elif isinstance(f, bson.DBRef):
            return
        elif is_dict and isinstance(f.get('_id'), bson.ObjectId):  # manual ref
            doc[updater.field_name] = bson.DBRef(collection_name, f['_id'])
        elif isinstance(f, bson.ObjectId):
            doc[updater.field_name] = bson.DBRef(collection_name, f)
        elif updater.migration_policy.name == 'strict':  # Other data type
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be DBRef, ObjectId, manual ref, dynamic ref) "
                                     f"in record {doc}")

    # TODO: precheck if field actually contains value other than dynamic ref or DBRef
    updater.update_by_document(by_doc)


def to_string(updater: DocumentUpdater):
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        is_dict = isinstance(f, dict)
        if is_dict and isinstance(f.get('_ref'), bson.DBRef):  # dynamic ref
            doc[updater.field_name] = str(f['_ref'].id)
        elif is_dict and isinstance(f.get('_id'), bson.ObjectId):  # manual ref
            doc[updater.field_name] = str(f['_id'])
        elif isinstance(f, bson.DBRef):
            doc[updater.field_name] = str(f.id)
        else:
            try:
                doc[updater.field_name] = str(f)
            except (TypeError, ValueError) as e:
                if updater.migration_policy.name == 'strict':
                    raise MigrationError(f'Cannot convert value {updater.field_name}: '
                                         f'{doc[updater.field_name]} to string') from e

    # TODO: precheck if field actually contains value other than string
    updater.update_by_document(by_doc)


def to_int(updater: DocumentUpdater):
    __mongo_convert(updater, 'int')


def to_long(updater: DocumentUpdater):
    __mongo_convert(updater, 'long')


def to_double(updater: DocumentUpdater):
    __mongo_convert(updater, 'double')


def to_decimal(updater: DocumentUpdater):
    __mongo_convert(updater, 'decimal')


def to_date(updater: DocumentUpdater):
    __mongo_convert(updater, 'date')


def to_bool(updater: DocumentUpdater):
    __mongo_convert(updater, 'bool')


def to_binary(updater: DocumentUpdater):
    __mongo_convert(updater, 'binary')


def to_object(updater: DocumentUpdater):
    __mongo_convert(updater, 'object')


def to_uuid_str(updater: DocumentUpdater):
    """Convert binData with UUID to string with UUID"""
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        if isinstance(f, str) and uuid_pattern.match(f):
            return
        elif isinstance(f, uuid.UUID):
            doc[updater.field_name] = str(f)
        elif updater.migration_policy.name == 'strict':
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be UUID string or UUID Binary data) in record {doc}")

    uuid_pattern = re.compile(r'\A[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}\Z',
                              re.IGNORECASE)
    updater.update_by_document(by_doc)


def to_uuid_bin(updater: DocumentUpdater):
    """Convert strings with UUID to binData with UUID"""
    def by_doc(ctx: ByDocContext):
        doc = ctx.document
        if updater.field_name not in doc or doc[updater.field_name] is None:
            return

        f = doc[updater.field_name]
        if isinstance(f, uuid.UUID):
            return
        elif isinstance(f, str) and uuid_pattern.match(f):
            doc[updater.field_name] = uuid.UUID(f)
        elif updater.migration_policy.name == 'strict':
            raise InconsistencyError(f"Field {updater.field_name} has wrong value {f!r} "
                                     f"(should be UUID string or UUID Binary data) in record {doc}")

    uuid_pattern = re.compile(r'\A[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}\Z',
                              re.IGNORECASE)
    updater.update_by_document(by_doc)


def to_url_string(updater: DocumentUpdater, check_only=False):
    """Cast fields to string and then verify if they contain URLs"""
    def by_path(ctx: ByPathContext):
        fltr = {ctx.filter_dotpath: {'$not': url_regex, '$ne': None}, **ctx.extra_filter}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    to_string(updater)

    url_regex = re.compile(
        r"\A[A-Z]{3,}://[A-Z0-9\-._~:/?#\[\]@!$&'()*+,;%=]\Z",
        re.IGNORECASE
    )
    if updater.migration_policy.name == 'strict':
        updater.update_by_path(by_path)


def to_email_string(updater: DocumentUpdater):
    def by_path(ctx: ByPathContext):
        fltr = {ctx.filter_dotpath: {'$not': email_regex, '$ne': None}, **ctx.extra_filter}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    to_string(updater)

    email_regex = re.compile(
        r"\A.*\Z",  # TODO: insert email validation regex here
        re.IGNORECASE
    )
    if updater.migration_policy.name == 'strict':
        updater.update_by_path(by_path)


def to_complex_datetime(updater: DocumentUpdater):
    def by_path(ctx: ByPathContext):
        fltr = {ctx.filter_dotpath: {'$not': regex, '$ne': None}, **ctx.extra_filter}
        check_empty_result(ctx.collection, ctx.filter_dotpath, fltr)

    to_string(updater)

    # We should not know which separator is used, so use '.+'
    # Separator change is handled by appropriate field method
    regex = r'\A' + str('.+'.join([r"\d{4}"] + [r"\d{2}"] * 5 + [r"\d{6}"])) + r'\Z'
    if updater.migration_policy.name == 'strict':
        updater.update_by_path(by_path)


def __mongo_convert(updater: DocumentUpdater, target_type: str):
    """
    Convert field to a given type in a given collection. `target_type`
    contains MongoDB type name, such as 'string', 'decimal', etc.

    https://docs.mongodb.com/manual/reference/operator/aggregation/convert/
    :param updater: DocumentUpdater object
    :param target_type: MongoDB type name
    :return:
    """
    def by_doc(ctx: ByDocContext):
        # https://docs.mongodb.com/manual/reference/operator/aggregation/convert/
        type_map = {
            'double': float,
            'string': str,
            'objectId': bson.ObjectId,
            'bool': bool,
            'date': lambda x: dateutil_parse(str(x)),
            'int': int,
            'long': bson.Int64,
            'decimal': float,
            'binary': bson.Binary,
            'object': dict
        }
        assert target_type in type_map

        doc = ctx.document
        field_name = updater.field_name
        if field_name in doc:
            t = type_map[target_type]
            if not isinstance(doc[field_name], t) and doc[field_name] is not None:
                try:
                    doc[field_name] = type_map[target_type](doc[field_name])
                except (TypeError, ValueError) as e:
                    if updater.migration_policy.name == 'strict':
                        raise MigrationError(f'Cannot convert value '
                                             f'{field_name}: {doc[field_name]} to type {t}') from e

    updater.update_by_document(by_doc)



__all__ = [
    'import_module',
    'collect_models_schema',
    'MongoengineMigrate',
]

import functools
import importlib.util
import logging
import re
from datetime import timezone, datetime
from pathlib import Path
from types import ModuleType
from typing import Tuple, Iterable, Optional, Dict

import pymongo.database
import pymongo.errors
from bson import CodecOptions
from dictdiffer import patch, swap
from jinja2 import Environment
from mongoengine.base import _document_registry
from pymongo import MongoClient

import mongoengine_migrate.flags as runtime_flags
from mongoengine_migrate.actions.factory import build_actions_chain
from mongoengine_migrate.exceptions import MongoengineMigrateError, ActionError, MigrationGraphError
from mongoengine_migrate.fields.registry import type_key_registry
from mongoengine_migrate.graph import Migration, MigrationsGraph, MigrationPolicy
from mongoengine_migrate.query_tracer import DatabaseQueryTracer
from mongoengine_migrate.schema import Schema
from mongoengine_migrate.utils import get_closest_parent, get_document_type

log = logging.getLogger('mongoengine-migrate')


def symbol_wrap(value: str, width: int = 80, wrap_by: str = ',', wrapstring: str = '\n'):
    """
    Jinja2 filter which implements word wrap, but can split the string
    only by given character. If no character found then does nothing.
    :param value: Incoming value
    :param width: Maximum width of string. Default is 80
    :param wrap_by: Symbol which we're split by. Default is comma (,)
    :param wrapstring: String to join each wrapped line. Default is
     newline
    :return: Wrapped string
    """
    pos = 1
    lines = []
    content_width = width - len(wrapstring.replace('\n', ''))
    while pos:
        pos = value.rfind(wrap_by, 0, content_width) + 1 or None
        if len(value) <= content_width:
            lines.append(value)
            break
        else:
            lines.append(value[:pos])
        value = value[pos:].lstrip(' \t')

    return wrapstring.join(lines)


def import_module(path: str) -> Tuple[ModuleType, str]:
    """
    Import module by python path
    :param path: dot path
    :return: tuple with module object and path part inside it
    """
    attrs = []
    while 1:
        try:
            module = importlib.import_module(path)
            rest = '.'.join(reversed(attrs))
            return module, rest
        except ModuleNotFoundError as e:
            try:
                path, attr = path.rsplit('.', 1)
            except ValueError:
                raise e
            attrs.append(attr)


def collect_models_schema() -> Schema:
    """
    Transform all available mongoengine document objects to db schema
    :return:
    """
    schema = Schema()
    collections: Dict[str, set] = {}  # {collection_name: set(top_level_documents)}

    # Retrieve models from mongoengine global document registry
    for model_cls in _document_registry.values():
        log.debug('> Reading document %s', repr(model_cls))
        # NOTE: EmbeddedDocuments are not append 'abstract' in meta if
        # `meta` is defined
        if model_cls._meta.get('abstract'):
            log.debug('> Skip %s since it is an abstract document')
            continue

        document_type = get_document_type(model_cls)
        if document_type is None:
            raise ActionError(f'Could not get document type for {model_cls!r}')

        if document_type in schema:
            raise ActionError(f'Models with the same document types {document_type!r} found')

        schema[document_type] = Schema.Document()
        if not document_type.startswith(runtime_flags.EMBEDDED_DOCUMENT_NAME_PREFIX):
            col = schema[document_type].parameters['collection'] = model_cls._get_collection_name()

            # Determine if unrelated documents have the same collection
            # E.g. DropDocument could drop all of these documents
            top_lvl_doc = document_type.split(runtime_flags.DOCUMENT_NAME_SEPARATOR)[0]
            collections.setdefault(col, set())
            if collections[col] and top_lvl_doc not in collections[col]:
                collections[col].add(top_lvl_doc)
                log.warning(f'These mongoengine documents use the same collection '
                            f'which can be a cause of data corruption: {collections[col]}')
            else:
                collections[col].add(top_lvl_doc)

        if model_cls._meta.get('allow_inheritance'):
            schema[document_type].parameters['inherit'] = True

        if model_cls._dynamic:
            schema[document_type].parameters['dynamic'] = True

        # {field_cls: TypeKeyRegistryItem}
        field_mapping_registry = {x.field_cls: x for x in type_key_registry.values()}

        # Collect schema for every field
        for field_name, field_obj in model_cls._fields.items():
            # Exclude '_id' special MongoDB field since it is immutable
            # and should not be a part of migrations
            if field_obj.db_field == '_id':
                continue

            field_cls = field_obj.__class__

            if field_cls in field_mapping_registry:
                registry_field_cls = field_cls
            else:
                registry_field_cls = get_closest_parent(
                    field_cls,
                    field_mapping_registry.keys()
                )

            if registry_field_cls is None:
                raise ActionError(f'Could not find {field_cls!r} or one of its base classes '
                                  f'in type_key registry')

            handler_cls = field_mapping_registry[registry_field_cls].field_handler_cls
            schema[document_type][field_name] = handler_cls.build_schema(field_obj)
            # TODO: validate default against all field restrictions such as min_length, regex, etc.

        log.debug("> Schema '%s' => %s", document_type, str(schema[document_type]))

    return schema


class MongoengineMigrate:
    default_collection_name: str = 'mongoengine_migrate'
    default_directory: str = './migrations'
    default_models_module = 'models'

    def __init__(self, mongo_uri: str, collection_name: str, migrations_dir: str, **kwargs):
        self.mongo_uri = mongo_uri
        self.migrations_collection_name = collection_name
        self.migration_dir = migrations_dir
        self._kwargs = kwargs
        self.client = MongoClient(mongo_uri)
        # Another MongoClient for additional operations which should
        # be performed in separate connection such as parallel bulk
        # writes
        self.client2 = MongoClient(mongo_uri)

        # Initiate immediate connect to MongoDB in order to ensure
        # that it is accessible
        log.debug('Connecting to MongoDB...')
        self.client.get_database().command('ping')

        # Trying to figure out server version if not specified
        if runtime_flags.mongo_version is None:
            try:
                server_info = self.client.server_info()
                runtime_flags.mongo_version = server_info['version']
                log.info('MongoDB version: %s', runtime_flags.mongo_version)
            except pymongo.errors.OperationFailure as e:
                raise MongoengineMigrateError(
                    'Could not figure out MongoDB version. Please set up '
                    'right permissions to be able to execute "buildinfo" '
                    'command or use --mongo-version argument to set version by hand'
                ) from e

    @functools.cached_property
    def db(self) -> pymongo.database.Database:
        """Return MongoDB database object"""
        db = self.client.get_database()
        if runtime_flags.dry_run:
            log.debug('> Dry run mode requested, use mock database object for main connection')
            db = DatabaseQueryTracer(db)

        return db

    @functools.cached_property
    def db2(self) -> pymongo.database.Database:
        """Return MongoDB database object for client2"""
        db = self.client2.get_database()
        if runtime_flags.dry_run:
            log.debug('> Dry run mode requested, use mock database object for second connection')
            db = DatabaseQueryTracer(db)

        return db

    @property
    def migration_collection(self) -> pymongo.collection.Collection:
        """Return collection object where we keep migration data"""
        return self.client.get_database()[self.migrations_collection_name].with_options(
            codec_options=CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        )

    def get_db_migration_names(self) -> Iterable[str]:
        """
        Return iterable with migration names was written in db in
        applying order
        """
        fltr = {'type': 'migrations'}
        for migration in self.migration_collection.find(fltr).sort('ordering_number'):
            return [m['name'] for m in migration['value']]

        return []

    def write_db_migrations_graph(self, graph: MigrationsGraph):
        """
        Write migrations graph to db
        :param graph: migrations graph
        """
        fltr = {'type': 'migrations'}
        records = []
        num = 0
        for migration in graph.walk_down(graph.initial, False):
            if migration.applied:
                records.append({
                    'name': migration.name,
                    'ordering_number': num
                })
                num += 1

        data = {'type': 'migrations', 'value': records}
        self.migration_collection.replace_one(fltr, data, upsert=True)

    def load_db_schema(self) -> Schema:
        """Load schema from db"""
        fltr = {'type': 'schema'}
        res = self.migration_collection.find_one(fltr)
        schema = Schema()
        schema.load(res.get('value', {}) if res else {})
        return schema

    def write_db_schema(self, schema: Schema) -> None:
        """
        Write schema to db
        :param schema:
        :return:
        """
        fltr = {'type': 'schema'}
        data = {'type': 'schema', 'value': schema.dump()}
        self.migration_collection.replace_one(fltr, data, upsert=True)

    def load_migrations(self,
                        directory: Path,
                        namespace: str = f"{__name__}._migrations") -> Iterable[Migration]:
        """
        Load migrations python modules located in a given directory.
        Every module is loaded on a given namespace. Function skips
        modules which names are started from double underscore
        :param directory: directory where modules will be searhed for
        :param namespace: namespace name where module names will be
         loaded
        :return: unsorted iterable with Migration objects
        """
        # FIXME: do not load to current namespace
        if not directory.exists():
            raise MongoengineMigrateError(f"Directory '{directory}' does not exist")

        for module_file in directory.glob("*.py"):
            if module_file.name.startswith("__"):
                continue

            log.debug('> Loading migration file %s', module_file)
            migration_name = module_file.stem
            spec = importlib.util.spec_from_file_location(
                f"{namespace}.{migration_name}", str(module_file)
            )
            migration_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(migration_module)
            yield Migration(
                name=migration_name,
                module=migration_module,
                dependencies=migration_module.dependencies
            )

    def build_graph(self) -> MigrationsGraph:
        """Build migrations graph with all migration modules"""
        graph = MigrationsGraph()
        for m in self.load_migrations(Path(self.migration_dir)):
            graph.add(m)

        applied = []
        for migration_name in self.get_db_migration_names():
            if migration_name not in graph.migrations:
                raise MigrationGraphError(
                    f'Migration {migration_name} was applied, but its python module not found. '
                    f'You can use schema repair to fix this issue'
                )
            graph.migrations[migration_name].applied = True
            applied.append(migration_name)

        log.debug('> Applied migrations: %s', applied)
        if graph.last:
            log.debug('> Last migration is: %s', graph.last.name)
        else:
            log.debug('> Last migration is: %s', 'None')

        return graph

    def upgrade(self, migration_name: str, graph: Optional[MigrationsGraph] = None):
        """
        Upgrade db to the given migration
        :param migration_name: target migration name
        :param graph: Optional. Migrations graph. If omitted, then it
         will be loaded
        :return:
        """
        if graph is None:
            log.debug('Loading migration files...')
            graph = self.build_graph()
        log.debug('Loading schema from database...')
        left_schema = self.load_db_schema()

        if migration_name not in graph.migrations:
            raise MigrationGraphError(f'Migration {migration_name} not found')

        db = self.db
        for migration in graph.walk_down(graph.initial, unapplied_only=True):
            log.info('Upgrading %s...', migration.name)
            for idx, action_object in enumerate(migration.get_actions(), start=1):
                log.debug('> [%d] %s', idx, str(action_object))
                if not action_object.dummy_action and not runtime_flags.schema_only:
                    action_object.prepare(db, left_schema, migration.policy)
                    action_object.run_forward()
                    action_object.cleanup()

                try:
                    left_schema = patch(action_object.to_schema_patch(left_schema), left_schema)
                except (TypeError, ValueError, KeyError) as e:
                    raise ActionError(
                        f"Unable to apply schema patch of {action_object!r}. More likely that the "
                        f"schema is corrupted. You can use schema repair tools to fix this issue"
                    ) from e

            graph.migrations[migration.name].applied = True

            if not runtime_flags.dry_run:
                log.debug('Writing db schema and migrations graph...')
                self.write_db_schema(left_schema)
                self.write_db_migrations_graph(graph)

            if migration.name == migration_name:
                break   # We've reached the target migration

        self._verify_schema(left_schema)

    def downgrade(self, migration_name: str, graph: Optional[MigrationsGraph] = None):
        """
        Downgrade db to the given migration
        :param migration_name: target migration name
        :param graph: Optional. Migrations graph. If omitted, then it
         will be loaded
        :return:
        """
        if graph is None:
            log.debug('Loading migration files...')
            graph = self.build_graph()
        log.debug('Loading schema from database...')
        left_schema = self.load_db_schema()

        if migration_name not in graph.migrations:
            raise MigrationGraphError(f'Migration {migration_name} not found')

        log.debug('Precalculating schema diffs...')
        # Collect schema diffs across all migrations
        migration_diffs = {}  # {migration_name: [action1_diff, ...]}
        temp_left_schema = Schema()
        for migration in graph.walk_down(graph.initial, unapplied_only=False):
            migration_diffs[migration.name] = []
            for action in migration.get_actions():
                forward_patch = action.to_schema_patch(temp_left_schema)
                migration_diffs[migration.name].append(forward_patch)

                try:
                    temp_left_schema = patch(forward_patch, temp_left_schema)
                except (TypeError, ValueError, KeyError) as e:
                    raise ActionError(
                        f"Unable to apply schema patch of {action!r}. More likely that the "
                        f"schema is corrupted. You can use schema repair tools to fix this issue"
                    ) from e

        db = self.db
        for migration in graph.walk_up(graph.last, applied_only=True):
            if migration.name == migration_name:
                break  # We've reached the target migration

            log.info('Downgrading %s...', migration.name)

            action_diffs = zip(
                migration.get_actions(),
                migration_diffs[migration.name],
                range(1, len(migration.get_actions()) + 1)
            )
            for action_object, action_diff, idx in reversed(list(action_diffs)):
                log.debug('> [%d] %s', idx, str(action_object))

                try:
                    left_schema = patch(list(swap(action_diff)), left_schema)
                except (TypeError, ValueError, KeyError) as e:
                    raise ActionError(
                        f"Unable to apply schema patch of {action_object!r}. More likely that the "
                        f"schema is corrupted. You can use schema repair tools to fix this issue"
                    ) from e

                if not action_object.dummy_action and not runtime_flags.schema_only:
                    action_object.prepare(db, left_schema, migration.policy)
                    action_object.run_backward()
                    action_object.cleanup()

            graph.migrations[migration.name].applied = False

            if not runtime_flags.dry_run:
                log.debug('Writing db schema and migrations graph...')
                self.write_db_schema(left_schema)
                self.write_db_migrations_graph(graph)

        self._verify_schema(left_schema)

    def migrate(self, migration_name: str = None):
        """
        Migrate db in order to reach a given migration. This process
        may require either upgrading or downgrading
        :param migration_name: target migration name
        :return:
        """
        log.debug('Loading migration files...')
        graph = self.build_graph()
        if not graph.last:
            raise MigrationGraphError('No migrations found')

        if migration_name is None:
            migration_name = graph.last.name

        if migration_name not in graph.migrations:
            raise MigrationGraphError(f'Migration {migration_name} not found')

        migration = graph.migrations[migration_name]
        if migration.applied:
            self.downgrade(migration_name, graph)
        else:
            self.upgrade(migration_name, graph)

    def makemigrations(self):
        """
        Compare current mongoengine documents state and the last db
        state and make a migration file if needed
        """
        log.debug('Loading migration files...')
        graph = self.build_graph()
        log.debug('Loading schema from database...')
        db_schema = self.load_db_schema()

        # Obtain schema changes which migrations would make (including
        #  unapplied ones)
        # If mongoengine models schema was changed regarding db schema
        #  then try to guess which actions would reflect such changes
        for migration in graph.walk_down(graph.initial, unapplied_only=False):
            for action_object in migration.get_actions():
                try:
                    db_schema = patch(action_object.to_schema_patch(db_schema), db_schema)
                except (TypeError, ValueError, KeyError) as e:
                    raise ActionError(
                        f"Unable to apply schema patch of {action_object!r}. More likely that the "
                        f"schema is corrupted. You can use schema repair tools to fix this issue"
                    ) from e

        log.debug('Collecting schema from mongoengine documents...')
        models_schema = collect_models_schema()
        if db_schema == models_schema:
            log.info('No changes detected')
            return

        log.debug('Building actions chain...')
        actions_chain = build_actions_chain(db_schema, models_schema)

        import_expressions = {'from mongoengine_migrate.actions import *'}
        for action in actions_chain:
            # If `regex` is set in action, then we probably need 're'
            if isinstance(action.parameters.get('regex'), re.Pattern):
                import_expressions.add('import re')
                break

        log.debug('Writing migrations file...')
        env = Environment()
        env.filters['symbol_wrap'] = symbol_wrap
        tpl_ctx = {
            'graph': graph,
            'actions_chain': actions_chain,
            'policy_enum': MigrationPolicy,
            'import_expressions': import_expressions
        }
        tpl_path = Path(__file__).parent / 'migration_template.tpl'
        tpl = env.from_string(tpl_path.read_text())
        migration_source = tpl.render(tpl_ctx)

        seq_number = str(len(graph.migrations)).zfill(4)
        name = f'{seq_number}_auto_{datetime.now().strftime("%Y%m%d_%H%M")}.py'
        migration_file = Path(self.migration_dir) / name
        migration_file.write_text(migration_source)

        log.info('Migration file "%s" was created', migration_file)

    def _verify_schema(self, schema: Schema):
        # Check if all derived documents have the same collection as
        # their parents.
        # E.g. user could comment/remove AlterDocument(collection=...)
        # for any derived document, but leave for base one)
        collections = {}  # {top_level_document: collection}
        for document_type, doc_schema in schema.items():
            if 'collection' not in doc_schema.parameters:
                continue

            col = doc_schema.parameters['collection']
            top_lvl_doc = document_type.split(runtime_flags.DOCUMENT_NAME_SEPARATOR)[0]
            if top_lvl_doc in collections and collections[top_lvl_doc] != col:
                log.warning(f'The collection in derived document {document_type} ({col}) '
                            f'is differ than its base document {top_lvl_doc} '
                            f'({collections[top_lvl_doc]}). Please fix collection name and rerun '
                            f'an affected migration')
            collections.setdefault(top_lvl_doc, col)



from copy import deepcopy

import pytest
import jsonpath_rw
import itertools

from mongoengine_migrate.actions import DropField
from mongoengine_migrate.exceptions import SchemaError
from mongoengine_migrate.graph import MigrationPolicy


class TestDropFieldInDocument:
    def test_forward__should_drop_field(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        expect = dump_db()
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for rec in parser.find(expect):
            if 'doc1_str' in rec.value:
                del rec.value['doc1_str']

        action = DropField('Schema1Doc1', 'doc1_str')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert expect == dump_db()

    def test_backward__if_default_is_not_set__should_do_nothing(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = DropField('Schema1Doc1', 'doc1_str')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump == dump_db()

    def test_backward__if_required_and_default_is_set__should_create_field_and_set_a_value(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        default = 'test!'
        schema['Schema1Doc1']['test_field'] = {
            'unique_with': None,
            'null': False,
            'choices': None,
            'default': default,
            'sparse': False,
            'unique': False,
            'required': True,
            'db_field': 'test_field',
            'primary_key': False,
            'type_key': 'StringField',
            'max_length': None,
            'regex': None,
            'min_length': None
        }
        dump = dump_db()
        expect = deepcopy(dump)
        parser = jsonpath_rw.parse('schema1_doc1[*]')
        for rec in parser.find(expect):
            rec.value['test_field'] = default

        action = DropField('Schema1Doc1', 'test_field')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()
        assert expect == dump_db()

    def test_backward__if_required_and_default_is_set_and_field_in_db__should_not_touch_field(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        default = 'test!'
        schema['Schema1Doc1']['test_field'] = {
            'unique_with': None,
            'null': False,
            'choices': None,
            'default': default,
            'sparse': False,
            'unique': False,
            'required': True,
            'db_field': 'test_field',
            'primary_key': False,
            'type_key': 'StringField',
            'max_length': None,
            'regex': None,
            'min_length': None
        }
        ids = set()
        for doc in test_db['schema1_doc1'].find({}, limit=2):
            test_db['schema1_doc1'].update_one({'_id': doc['_id']},
                                               {'$set': {'test_field': 'old_value'}})
            ids.add(doc['_id'])

        action = DropField('Schema1Doc1', 'test_field')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert all(d['test_field'] == 'old_value'
                   for d in test_db['schema1_doc1'].find()
                   if d['_id'] in ids)

    def test_prepare__if_such_document_is_not_in_schema__should_raise_error(self,
                                                                            load_fixture,
                                                                            test_db):
        schema = load_fixture('schema1').get_schema()
        del schema['Schema1Doc1']

        action = DropField('Schema1Doc1', 'doc1_str')

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)

    def test_prepare__if_such_field_in_document_is_not_in_schema__should_raise_error(self,
                                                                                     load_fixture,
                                                                                     test_db):
        schema = load_fixture('schema1').get_schema()

        action = DropField('Schema1Doc1', 'unknown_field')

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)


class TestCreateFieldEmbedded:
    def test_backward__if_default_is_not_set__should_do_nothing(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = DropField('~Schema1EmbDoc1', 'embdoc1_str')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump == dump_db()

    def test_backward__if_required_and_default_is_set__should_create_field_and_set_a_value(
            self, load_fixture, test_db, dump_db
    ):
        schema = load_fixture('schema1').get_schema()
        default = 'test!'
        schema['~Schema1EmbDoc1']['test_field'] = {
            'unique_with': None,
            'null': False,
            'choices': None,
            'default': default,
            'sparse': False,
            'unique': False,
            'required': True,
            'db_field': 'test_field',
            'primary_key': False,
            'type_key': 'StringField',
            'max_length': None,
            'regex': None,
            'min_length': None
        }
        dump = dump_db()
        expect = deepcopy(dump)
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for rec in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            rec.value['test_field'] = default

        action = DropField('~Schema1EmbDoc1', 'test_field')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert expect == dump_db()

    def test_forward__should_drop_field(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()
        expect = deepcopy(dump)
        parsers = load_fixture('schema1').get_embedded_jsonpath_parsers('~Schema1EmbDoc1')
        for rec in itertools.chain.from_iterable(p.find(expect) for p in parsers):
            if 'embdoc1_str' in rec.value:
                del rec.value['embdoc1_str']

        action = DropField('~Schema1EmbDoc1', 'embdoc1_str')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert expect == dump_db()



from copy import deepcopy

import pytest

from mongoengine_migrate.actions import DropDocument
from mongoengine_migrate.exceptions import SchemaError
from mongoengine_migrate.graph import MigrationPolicy


class TestDropDocument:
    def test_forward__should_drop_collection(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = DropDocument('Schema1Doc1')
        action.prepare(test_db, schema, MigrationPolicy.strict)
        expect = deepcopy(dump)
        del expect['schema1_doc1']

        action.run_forward()

        assert expect == dump_db()

    def test_forward__on_unexistance_collection_specified__should_do_nothing(self,
                                                                             load_fixture,
                                                                             test_db,
                                                                             dump_db):
        schema = load_fixture('schema1').get_schema()
        schema['Schema1Doc1'].parameters['collection'] = 'unknown_collection'
        dump = dump_db()

        action = DropDocument('Schema1Doc1')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_forward()

        assert dump == dump_db()

    def test_backward__should_do_nothing(self, load_fixture, test_db, dump_db):
        schema = load_fixture('schema1').get_schema()
        dump = dump_db()

        action = DropDocument('Schema1Doc1')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump == dump_db()

    def test_backward__on_unexistance_collection_specified__should_do_nothing(self,
                                                                              load_fixture,
                                                                              test_db,
                                                                              dump_db):
        schema = load_fixture('schema1').get_schema()
        schema['Schema1Doc1'].parameters['collection'] = 'unknown_collection'
        dump = dump_db()

        action = DropDocument('Schema1Doc1')
        action.prepare(test_db, schema, MigrationPolicy.strict)

        action.run_backward()

        assert dump == dump_db()

    def test_prepare__if_such_document_is_not_in_schema__should_raise_error(self,
                                                                            load_fixture,
                                                                            test_db):
        schema = load_fixture('schema1').get_schema()
        del schema['Schema1Doc1']

        action = DropDocument('Schema1Doc1')

        with pytest.raises(SchemaError):
            action.prepare(test_db, schema, MigrationPolicy.strict)
